"""Test that we don't fail the same test cases that Beautiful Soup used to fail prior to Soup Sieve."""
from bs4 import BeautifulSoup
import unittest
import soupsieve as sv
from . import util


class SelectorNthOfTypeBugTest(unittest.TestCase):
    """
    Original Beautiful soup test html document.

    http://bazaar.launchpad.net/~leonardr/beautifulsoup/bs4/view/head:/bs4/tests/test_tree.py, line 1627.
    """

    HTML = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
"http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
<title>The title</title>
<link rel="stylesheet" href="blah.css" type="text/css" id="l1">
</head>
<body>
<custom-dashed-tag class="dashed" id="dash1">Hello there.</custom-dashed-tag>
<div id="main" class="fancy">
<div id="inner">
<h1 id="header1">An H1</h1>
<p>Some text</p>
<p class="onep" id="p1">Some more text</p>
<h2 id="header2">An H2</h2>
<p class="class1 class2 class3" id="pmulti">Another</p>
<a href="http://bob.example.org/" rel="friend met" id="bob">Bob</a>
<h2 id="header3">Another H2</h2>
<a id="me" href="http://simonwillison.net/" rel="me">me</a>
<span class="s1">
<a href="#" id="s1a1">span1a1</a>
<a href="#" id="s1a2">span1a2 <span id="s1a2s1">test</span></a>
<span class="span2">
<a href="#" id="s2a1">span2a1</a>
</span>
<span class="span3"></span>
<custom-dashed-tag class="dashed" id="dash2"/>
<div data-tag="dashedvalue" id="data1"/>
</span>
</div>
<x id="xid">
<z id="zida"/>
<z id="zidab"/>
<z id="zidac"/>
</x>
<y id="yid">
<z id="zidb"/>
</y>
<p lang="en" id="lang-en">English</p>
<p lang="en-gb" id="lang-en-gb">English UK</p>
<p lang="en-us" id="lang-en-us">English US</p>
<p lang="fr" id="lang-fr">French</p>
</div>

<div id="footer">
</div>
"""

    def setUp(self):
        """Setup."""

        self.soup = BeautifulSoup(self.HTML, 'html.parser')

    def test_parent_nth_of_type_preconditions(self):
        """Test `nth` type preconditions."""

        els = sv.select('div > h1', self.soup)
        # check that there is a unique selection
        self.assertEqual(len(els), 1)
        self.assertEqual(els[0].string, 'An H1')

        # Show that the `h1`'s parent `div#inner` is the first child of type `div` of the grandparent `div#main`.
        # so that the selector `div:nth-of-type(1) > h1` should also give `h1`.
        h1 = els[0]
        div_inner = h1.parent
        div_main = div_inner.parent
        div_main_children = [child for child in div_main.children]
        self.assertEqual(div_main_children[0], '\n')
        self.assertEqual(div_main_children[1], div_inner)

    def test_parent_nth_of_type(self):
        """Test parent of `nth` of type."""

        els = sv.select('div:nth-of-type(1) > h1', self.soup)
        self.assertEqual(len(els), 1)
        self.assertEqual(els[0].string, 'An H1')


SIMPLE_XML = """<Envelope><Header>...</Header></Envelope>"""
NAMESPACE_XML = """
<?xml version="1.0"?>
<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://www.w3.org/2005/08/addressing"
            xmlns:u="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
  <a:Action s:mustUnderstand="1">http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue</a:Action>
  <o:UsernameToken u:Id="uuid-00000043-0000-4000-0000-000000000000">
</s:Envelope>
""".strip()
NAMESPACES = dict(x="http://www.w3.org/2003/05/soap-envelope",
                  y="http://www.w3.org/2005/08/addressing",
                  z="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd")


@util.requires_lxml
def test_simple_xml():
    """Test basic XML."""
    xml = BeautifulSoup(SIMPLE_XML, "xml")

    assert xml.select_one("Envelope")
    assert xml.select_one("Envelope Header")
    assert xml.select_one("Header")

    assert not xml.select_one("envelope")
    assert not xml.select_one("envelope header")
    assert not xml.select_one("header")


@util.requires_lxml
def test_namespace_xml():
    """Test namespace XML."""
    xml = BeautifulSoup(NAMESPACE_XML, "xml")

    assert xml.select_one("Envelope")
    assert xml.select_one("Envelope Action")
    assert xml.select_one("Action")

    assert not xml.select_one("envelope")
    assert not xml.select_one("envelope action")
    assert not xml.select_one("action")


@util.requires_lxml
def test_namespace_xml_with_namespace():
    """Test namespace selectors with XML."""
    xml = BeautifulSoup(NAMESPACE_XML, "xml")

    assert xml.select_one("x|Envelope", namespaces=NAMESPACES)
    assert xml.select_one("x|Envelope y|Action", namespaces=NAMESPACES)
    assert xml.select_one("y|Action", namespaces=NAMESPACES)

    assert not xml.select_one("x|envelope", namespaces=NAMESPACES)
    assert not xml.select_one("x|envelope y|action", namespaces=NAMESPACES)
    assert not xml.select_one("y|action", namespaces=NAMESPACES)



"""Test child combinators."""
from .. import util
from soupsieve import SelectorSyntaxError


class TestChild(util.TestCase):
    """Test child combinators."""

    MARKUP = """
    <div>
    <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3">Direct child</span>
    <pre>
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    """

    def test_direct_child(self):
        """Test direct child."""

        # Spaces
        self.assert_selector(
            self.MARKUP,
            "div > span",
            ["3"],
            flags=util.HTML
        )

    def test_direct_child_no_spaces(self):
        """Test direct child with no spaces."""

        # No spaces
        self.assert_selector(
            self.MARKUP,
            "div>span",
            ["3"],
            flags=util.HTML
        )

    def test_invalid_double_combinator(self):
        """Test that selectors cannot have double combinators."""

        self.assert_raises('div >> p', SelectorSyntaxError)
        self.assert_raises('>> div > p', SelectorSyntaxError)

    def test_invalid_trailing_combinator(self):
        """Test that selectors cannot have a trailing combinator."""

        self.assert_raises('div >', SelectorSyntaxError)
        self.assert_raises('div >, div', SelectorSyntaxError)

    def test_invalid_combinator(self):
        """Test that we do not allow selectors in selector lists to start with combinators."""

        self.assert_raises('> p', SelectorSyntaxError)
        self.assert_raises('div, > a', SelectorSyntaxError)



"""Test link selectors."""
from .. import util


class TestLink(util.TestCase):
    """Test link selectors."""

    MARKUP = """
    <div>
    <p>Some text <span id="1" class="foo:bar:foobar"> in a paragraph</span>.
    <a id="2" class="bar" href="http://google.com">Link</a>
    <a id="3">Placeholder text.</a>
    </p>
    </div>
    """

    def test_link(self):
        """Test link (all links are unvisited)."""

        self.assert_selector(
            self.MARKUP,
            ":link",
            ["2"],
            flags=util.HTML
        )

        self.assert_selector(
            self.MARKUP,
            r":\liNk",
            ["2"],
            flags=util.HTML
        )

    def test_tag_and_link(self):
        """Test link and tag (all links are unvisited)."""

        self.assert_selector(
            self.MARKUP,
            "a:link",
            [],
            flags=util.XML
        )



"""Test attribute selector."""
from .. import util
from soupsieve import SelectorSyntaxError
from bs4 import BeautifulSoup as BS


class TestAttribute(util.TestCase):
    """Test attribute selector."""

    MARKUP = """
    <div id="div">
    <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3">Direct child</span>
    <pre id="pre">
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    """

    MARKUP_CONTAINS = """
    <div id="div">
    <p id="0" class="test1 test2 test3">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3">Direct child</span>
    <pre id="pre" class="test-a test-b">
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    """

    # Browsers normally replace NULL with `\uFFFD`, but some of the parsers
    # we test just strip out NULL, so we will simulate and just insert `\uFFFD` directly
    # to ensure consistent behavior in our tests across parsers.
    MARKUP_NULL = """
    <div id="div">
    <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3">Direct child</span>
    <pre id="\ufffdpre">
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    """

    def test_attribute(self):
        """Test attribute."""

        self.assert_selector(
            self.MARKUP,
            "[href]",
            ["2"],
            flags=util.HTML
        )

    def test_attribute_with_spaces(self):
        """Test attribute with spaces."""

        # With spaces
        self.assert_selector(
            self.MARKUP,
            "[   href   ]",
            ["2"],
            flags=util.HTML
        )

    def test_multi_attribute(self):
        """Test multiple attribute."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4" class="test">Child 1</span>
            <span id="5" class="test" data-test="test">Child 2</span>
            <span id="6">Child 3</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "span[id].test[data-test=test]",
            ["5"],
            flags=util.HTML
        )

    def test_attribute_equal_no_quotes(self):
        """Test attribute with value that equals specified value (with no quotes)."""

        # No quotes
        self.assert_selector(
            self.MARKUP,
            '[id=\\35]',
            ["5"],
            flags=util.HTML
        )

    def test_attribute_equal_with_quotes(self):
        """Test attribute with value that equals specified value (with quotes)."""

        # Single quoted
        self.assert_selector(
            self.MARKUP,
            "[id='5']",
            ["5"],
            flags=util.HTML
        )

    def test_attribute_equal_with_double_quotes(self):
        """Test attribute with value that equals specified value (with double quotes)."""

        # Double quoted
        self.assert_selector(
            self.MARKUP,
            '[id="5"]',
            ["5"],
            flags=util.HTML
        )

    def test_attribute_equal_quotes_and_spaces(self):
        """Test attribute with value that equals specified value (quotes and spaces)."""

        # With spaces
        self.assert_selector(
            self.MARKUP,
            '[  id  =  "5"  ]',
            ["5"],
            flags=util.HTML
        )

    def test_attribute_equal_case_insensitive_attribute(self):
        """Test attribute with value that equals specified value (case insensitive attribute)."""

        self.assert_selector(
            self.MARKUP,
            '[ID="5"]',
            ["5"],
            flags=util.HTML
        )

    def test_attribute_bad(self):
        """Test attribute with a bad attribute."""

        self.assert_selector(
            '<span bad="5"></span>',
            '[  id  =  "5"  ]',
            [],
            flags=util.HTML
        )

    def test_attribute_escaped_newline(self):
        """Test attribute with escaped new line in quoted string."""

        self.assert_selector(
            self.MARKUP,
            '[id="pr\\\ne"]',
            ["pre"],
            flags=util.HTML
        )

    def test_attribute_equal_literal_null(self):
        """Test attribute with value that equals specified value with a literal null character."""

        self.assert_selector(
            self.MARKUP_NULL,
            '[id="\x00pre"]',
            ["\ufffdpre"],
            flags=util.HTML
        )

    def test_attribute_equal_escaped_null(self):
        """Test attribute with value that equals specified value with an escaped null character."""

        self.assert_selector(
            self.MARKUP_NULL,
            r'[id="\0 pre"]',
            ["\ufffdpre"],
            flags=util.HTML
        )

    def test_invalid_tag(self):
        """
        Test invalid tag.

        Tag must come first.
        """

        self.assert_raises('[href]p', SelectorSyntaxError)

    def test_malformed(self):
        """Test malformed."""

        # Malformed attribute
        self.assert_raises('div[attr={}]', SelectorSyntaxError)

    def test_attribute_type_html(self):
        """Type is treated as case insensitive in HTML."""

        markup = """
        <html>
        <body>
        <div id="div">
        <p type="TEST" id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a type="test" id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre id="pre">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        </body>
        </html>
        """

        self.assert_selector(
            markup,
            '[type="test"]',
            ["0", '2'],
            flags=util.HTML
        )

    def test_attribute_type_xml(self):
        """Type is treated as case sensitive in XML."""

        markup = """
        <html>
        <body>
        <div id="div">
        <p type="TEST" id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a type="test" id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre id="pre">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        </body>
        </html>
        """

        self.assert_selector(
            markup,
            '[type="test"]',
            ['2'],
            flags=util.XML
        )

    def test_attribute_type_xhtml(self):
        """Type is treated as case insensitive in XHTML."""

        markup = """
        <?xml version="1.0" encoding="UTF-8"?>
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
            "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
        <html lang="en" xmlns="http://www.w3.org/1999/xhtml">
        <head>
        </head>
        <body>
        <div id="div">
        <p type="TEST" id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a type="test" id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre id="pre">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        </body>
        </html>
        """

        self.assert_selector(
            markup,
            '[type="test"]',
            ['2'],
            flags=util.XHTML
        )

    def test_attribute_start_dash(self):
        """Test attribute whose dash separated value starts with the specified value."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0" lang="en-us">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "[lang|=en]",
            ["0"],
            flags=util.HTML
        )

    def test_attribute_contains_space_middle(self):
        """Test attribute whose space separated list contains the specified value in the middle of the list."""

        # Middle of list
        self.assert_selector(
            self.MARKUP_CONTAINS,
            "[class~=test2]",
            ["0"],
            flags=util.HTML
        )

    def test_attribute_contains_space_start(self):
        """Test attribute whose space separated list contains the specified value at the start of the list."""

        # Start of list
        self.assert_selector(
            self.MARKUP_CONTAINS,
            "[class~=test-a]",
            ["pre"],
            flags=util.HTML
        )

    def test_attribute_contains_space_end(self):
        """Test attribute whose space separated list contains the specified value at the end of the list."""

        # End of list
        self.assert_selector(
            self.MARKUP_CONTAINS,
            "[class~=test-b]",
            ["pre"],
            flags=util.HTML
        )

    def test_attribute_contains_cannot_have_spaces(self):
        """Test attribute `~=` will match nothing when spaces are included."""

        # Shouldn't match anything
        self.assert_selector(
            self.MARKUP_CONTAINS,
            '[class~="test1 test2"]',
            [],
            flags=util.HTML
        )

    def test_attribute_contains_cannot_have_empty(self):
        """Test attribute `~=` will match nothing when value is empty."""

        self.assert_selector(
            self.MARKUP_CONTAINS,
            '[class~=""]',
            [],
            flags=util.HTML
        )

    def test_attribute_contains_cannot_have_escaped_spaces(self):
        """Test attribute `~=` will match nothing when escaped spaces are included."""

        self.assert_selector(
            self.MARKUP_CONTAINS,
            '[class~="test1\\ test2"]',
            [],
            flags=util.HTML
        )

    def test_none_inputs(self):
        """Test weird inputs."""

        soup = BS('<span>text</span>', 'html.parser')
        soup.span['foo'] = None
        self.assertEqual(len(soup.select('span[foo]')), 1)

    def test_numeric_inputs(self):
        """Test weird inputs."""

        soup = BS('<span>text</span>', 'html.parser')
        soup.span['foo'] = 3
        self.assertEqual(len(soup.select('span[foo="3"]')), 1)
        soup.span['foo'] = 3.3
        self.assertEqual(len(soup.select('span[foo="3.3"]')), 1)

    def test_sequence_inputs(self):
        """Test weird inputs."""

        soup = BS('<span>text</span>', 'html.parser')
        soup.span['foo'] = [3, "4"]
        self.assertEqual(len(soup.select('span[foo="3 4"]')), 1)

    def test_bytes_inputs(self):
        """Test weird inputs."""

        soup = BS('<span>text</span>', 'html.parser')
        soup.span['foo'] = b'test'
        self.assertEqual(len(soup.select('span[foo="test"]')), 1)

    def test_weird_inputs(self):
        """Test weird inputs."""

        soup = BS('<span>text</span>', 'html.parser')
        soup.span['foo'] = {'3': '4'}
        self.assertEqual(len(soup.select('span[foo="{\'3\': \'4\'}"]')), 1)

    def test_nested_sequences(self):
        """Nested sequences will crash and burn due to the way BS handles them."""

        # The exact exception is not important as it can fail in various locations for different reasons
        with self.assertRaises(Exception):
            soup = BS('<span>text</span>', 'html.parser')
            soup.span['foo'] = [['1']]
            soup.select("span['foo']")



"""Test target within selectors."""
from .. import util


class TestTargetWithin(util.TestCase):
    """Test target within selectors."""

    MARKUP = """
    <a href="#head-2">Jump</a>
    <article id="article">
    <h2 id="head-1">Header 1</h1>
    <div><p>content</p></div>
    <h2 id="head-2">Header 2</h1>
    <div><p>content</p></div>
    </article>
    """

    def test_target_within(self):
        """Test target within."""

        self.assert_selector(
            self.MARKUP,
            "article:target-within",
            [],
            flags=util.HTML
        )

    def test_not_target_within(self):
        """Test inverse of target within."""

        self.assert_selector(
            self.MARKUP,
            "article:not(:target-within)",
            ["article"],
            flags=util.HTML
        )



"""Test custom selector aliases."""
from .. import util
import soupsieve as sv


class TestCustomSelectors(util.TestCase):
    """Test custom selector aliases."""

    MARKUP = """
    <body>
    <h1 id="1">Header 1</h1>
    <h2 id="2">Header 2</h2>
    <p id="3"></p>
    <p id="4"><span>child</span></p>
    </body>
    """

    def test_custom_selectors(self):
        """Test custom selectors."""

        custom_selectors = {
            ":--headers": "h1, h2, h3, h4, h5, h6",
            ":--parent": ":has(> *|*)"
        }

        self.assert_selector(
            self.MARKUP,
            ':--headers',
            ['1', '2'],
            custom=custom_selectors,
            flags=util.HTML
        )

        self.assert_selector(
            self.MARKUP,
            ':--headers:nth-child(2)',
            ['2'],
            custom=custom_selectors,
            flags=util.HTML
        )

        self.assert_selector(
            self.MARKUP,
            'p:--parent',
            ['4'],
            custom=custom_selectors,
            flags=util.HTML
        )

    def test_custom_escapes(self):
        """Test custom selectors with escapes."""

        custom_selectors = {
            r":--Header\s": "h1, h2, h3, h4, h5, h6"
        }

        self.assert_selector(
            self.MARKUP,
            r':--\HeaderS',
            ['1', '2'],
            custom=custom_selectors,
            flags=util.HTML
        )

    def test_custom_selectors_exotic(self):
        """Test custom selectors."""

        custom_selectors = {":--": "h1, h2, h3, h4, h5, h6"}

        self.assert_selector(
            self.MARKUP,
            ':--',
            ['1', '2'],
            custom=custom_selectors,
            flags=util.HTML
        )

    def test_custom_dependency(self):
        """Test custom selector dependency on other custom selectors."""

        custom_selectors = {
            ":--parent": ":has(> *|*)",
            ":--parent-paragraph": "p:--parent"
        }

        self.assert_selector(
            self.MARKUP,
            ':--parent-paragraph',
            ['4'],
            custom=custom_selectors,
            flags=util.HTML
        )

    def test_custom_dependency_out_of_order(self):
        """Test custom selector out of order dependency."""

        custom_selectors = {
            ":--parent-paragraph": "p:--parent",
            ":--parent": ":has(> *|*)"
        }

        self.assert_selector(
            self.MARKUP,
            ':--parent-paragraph',
            ['4'],
            custom=custom_selectors,
            flags=util.HTML
        )

    def test_custom_dependency_recursion(self):
        """Test that we fail on dependency recursion."""

        custom_selectors = {
            ":--parent-paragraph": "p:--parent",
            ":--parent": "p:--parent-paragraph"
        }

        self.assert_raises(':--parent', sv.SelectorSyntaxError, custom=custom_selectors)

    def test_bad_custom(self):
        """Test that a bad custom raises a syntax error."""

        custom_selectors = {
            ":--parent": ":has(> *|*)",
            ":--parent-paragraph": "p:--parent"
        }

        self.assert_raises(':--wrong', sv.SelectorSyntaxError, custom=custom_selectors)

    def test_bad_custom_syntax(self):
        """Test that a custom selector with bad syntax in its name fails."""

        self.assert_raises('div', sv.SelectorSyntaxError, custom={":--parent.": ":has(> *|*)"})

    def test_pseudo_class_collision(self):
        """Test that a custom selector cannot match an already existing pseudo-class name."""

        self.assert_raises('div', sv.SelectorSyntaxError, custom={":hover": ":has(> *|*)"})

    def test_custom_collision(self):
        """Test that a custom selector cannot match an already existing custom name."""

        self.assert_raises('div', KeyError, custom={":--parent": ":has(> *|*)", ":--PARENT": ":has(> *|*)"})



"""Test target selectors."""
from .. import util


class TestTarget(util.TestCase):
    """Test target selectors."""

    MARKUP = """
    <div>
    <h2 id="head-1">Header 1</h1>
    <div><p>content</p></div>
    <h2 id="head-2">Header 2</h1>
    <div><p>content</p></div>
    </div>
    """

    def test_target(self):
        """Test target."""

        self.assert_selector(
            self.MARKUP,
            "#head-2:target",
            [],
            flags=util.HTML
        )

    def test_not_target(self):
        """Test not target."""

        self.assert_selector(
            self.MARKUP,
            "#head-2:not(:target)",
            ["head-2"],
            flags=util.HTML
        )



"""Test disabled selectors."""
from .. import util


class TestDisabled(util.TestCase):
    """Test disabled selectors."""

    MARKUP = """
    <body>
    <form action="#">
      <fieldset id='a' disabled>
        <legend>
          Simple fieldset <input type="radio" id="1" checked>
          <fieldset id='b' disabled>
            <legend>Simple fieldset <input type="radio" id="2" checked></legend>
            <input type="radio" id="3" checked>
            <label for="radio">radio</label>
          </fieldset>
        </legend>
        <fieldset id='c' disabled>
          <legend>Simple fieldset <input type="radio" id="4" checked></legend>
          <input type="radio" id="5" checked>
          <label for="radio">radio</label>
        </fieldset>
        <input type="radio" id="6" checked>
        <label for="radio">radio</label>
      </fieldset>
      <optgroup>
        <option id="7" disabled>option</option>
      </optgroup>
      <optgroup id="8" disabled>
        <option id="9">option</option>
      </optgroup>
    </form>
    </body>
    """

    MARKUP_NESTED = """
    <body>
    <form action="#">
      <fieldset id='a' disabled>
        <legend>
          Simple fieldset <input type="radio" id="1" checked>
          <fieldset id='b'>
            <legend>Simple fieldset <input type="radio" id="2" checked></legend>
            <input type="radio" id="3" checked>
            <label for="radio">radio</label>
          </fieldset>
        </legend>
        <fieldset id='c' disabled>
          <legend>Simple fieldset <input type="radio" id="4" checked></legend>
          <input type="radio" id="5" checked>
          <label for="radio">radio</label>
        </fieldset>
        <input type="radio" id="6" checked>
        <label for="radio">radio</label>
      </fieldset>
      <optgroup>
        <option id="7" disabled>option</option>
      </optgroup>
      <optgroup id="8" disabled>
        <option id="9">option</option>
      </optgroup>
    </form>
    </body>
    """

    def test_disabled_html5(self):
        """
        Test disabled for HTML5 parser.

        Form elements that have `disabled`.
        `option` that is child of disabled `optgroup`.
        Form elements that are children of a disabled `fieldset`, but not it's `legend`.
        """

        self.assert_selector(
            self.MARKUP,
            ":disabled",
            ['3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c'],
            flags=util.HTML5
        )

    def test_disabled_lxml(self):
        """
        Test disabled for `lxml` HTML parser.

        Form elements that have `disabled`.
        `option` that is child of disabled `optgroup`.
        Form elements that are children of a disabled `fieldset`, but not it's `legend`.
        """

        self.assert_selector(
            self.MARKUP,
            ":disabled",
            ['2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c'],
            flags=util.LXML_HTML
        )

    def test_disabled_python(self):
        """
        Test disabled for the built-in HTML parser.

        Form elements that have `disabled`.
        `option` that is child of disabled `optgroup`.
        Form elements that are children of a disabled `fieldset`, but not it's `legend`.
        """

        self.assert_selector(
            self.MARKUP,
            ":disabled",
            ['3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c'],
            flags=util.PYHTML
        )

    def test_disabled_with_nested_disabled_form_html5(self):
        """Test disabled (with nested disabled forms) in the built-in HTML parser."""

        self.assert_selector(
            self.MARKUP_NESTED,
            ":disabled",
            ['4', '5', '6', '7', '8', '9', 'a', 'c'],
            flags=util.HTML5
        )

    def test_disabled_with_nested_disabled_form_lxml(self):
        """Test disabled (with nested disabled forms) in the `lxml` HTML parser."""

        self.assert_selector(
            self.MARKUP_NESTED,
            ":disabled",
            ['2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c'],
            flags=util.LXML_HTML
        )

    def test_disabled_with_nested_disabled_form_python(self):
        """Test disabled (with nested disabled forms) the built-in HTML parser."""

        self.assert_selector(
            self.MARKUP_NESTED,
            ":disabled",
            ['4', '5', '6', '7', '8', '9', 'a', 'c'],
            flags=util.PYHTML
        )

    def test_disabled_with_nested_optgroup(self):
        """Test `:disabled` only selects `option` elements whose closest `optgroup` parent is disabled."""

        self.assert_selector(
            """
            <optgroup id="0" disabled>
               <option id="1"></option>
               <optgroup id="3">
                   <option id="4"></option>
               </optgroup>
            </optgroup>
            """,
            ":disabled",
            ['0', '1'],
            flags=util.HTML
        )



"""Test required selectors."""
from .. import util


class TestRequired(util.TestCase):
    """Test required selectors."""

    MARKUP = """
    <form>
    <input id="1" type="name" required>
    <input id="2" type="checkbox" required>
    <input id="3" type="email">
    <textarea id="4" name="name" cols="30" rows="10" required></textarea>
    <select id="5" name="nm" required>
        <!-- options -->
    </select>
    </form>
    """

    def test_required(self):
        """Test required."""

        self.assert_selector(
            self.MARKUP,
            ":required",
            ['1', '2', '4', '5'],
            flags=util.HTML
        )

    def test_specific_required(self):
        """Test specific required."""

        self.assert_selector(
            self.MARKUP,
            "input:required",
            ['1', '2'],
            flags=util.HTML
        )



"""Test scope selectors."""
from .. import util
import soupsieve as sv


class TestScope(util.TestCase):
    """Test scope selectors."""

    MARKUP = """
    <html id="root">
    <head>
    </head>
    <body>
    <div id="div">
    <p id="0" class="somewordshere">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3" class="herewords">Direct child</span>
    <pre id="pre" class="wordshere">
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    </body>
    </html>
    """

    def test_scope_is_root(self):
        """Test scope is the root when the a specific element is not the target of the select call."""

        # Scope is root when applied to a document node
        self.assert_selector(
            self.MARKUP,
            ":scope",
            ["root"],
            flags=util.HTML
        )

        self.assert_selector(
            self.MARKUP,
            ":scope > body > div",
            ["div"],
            flags=util.HTML
        )

    def test_scope_cannot_select_target(self):
        """Test that scope, the element which scope is called on, cannot be selected."""

        for parser in util.available_parsers(
                'html.parser', 'lxml', 'html5lib', 'xml'):
            soup = self.soup(self.MARKUP, parser)
            el = soup.html

            # Scope is the element we are applying the select to, and that element is never returned
            self.assertTrue(len(sv.select(':scope', el, flags=sv.DEBUG)) == 0)

    def test_scope_is_select_target(self):
        """Test that scope is the element which scope is called on."""

        for parser in util.available_parsers(
                'html.parser', 'lxml', 'html5lib', 'xml'):
            soup = self.soup(self.MARKUP, parser)
            el = soup.html

            # Scope here means the current element under select
            ids = []
            for el in sv.select(':scope div', el, flags=sv.DEBUG):
                ids.append(el.attrs['id'])
            self.assertEqual(sorted(ids), sorted(['div']))

            el = soup.body
            ids = []
            for el in sv.select(':scope div', el, flags=sv.DEBUG):
                ids.append(el.attrs['id'])
            self.assertEqual(sorted(ids), sorted(['div']))

            # `div` is the current element under select, and it has no `div` elements.
            el = soup.div
            ids = []
            for el in sv.select(':scope div', el, flags=sv.DEBUG):
                ids.append(el.attrs['id'])
            self.assertEqual(sorted(ids), sorted([]))

            # `div` does have an element with the class `.wordshere`
            ids = []
            for el in sv.select(':scope .wordshere', el, flags=sv.DEBUG):
                ids.append(el.attrs['id'])
            self.assertEqual(sorted(ids), sorted(['pre']))



"""Test last child selectors."""
from .. import util


class TestLastChild(util.TestCase):
    """Test last child selectors."""

    MARKUP = """
    <div id="div">
    <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3">Direct child</span>
    <pre id="pre">
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    """

    def test_last_child(self):
        """Test last child."""

        self.assert_selector(
            self.MARKUP,
            "span:last-child",
            ["1", "6"],
            flags=util.HTML
        )

    def test_last_child_case(self):
        """Test last child token's case insensitivity."""

        self.assert_selector(
            self.MARKUP,
            "span:LAST-CHILD",
            ["1", "6"],
            flags=util.HTML
        )



"""Test next sibling combinators."""
from .. import util


class TestNextSibling(util.TestCase):
    """Test next sibling combinators."""

    MARKUP = """
    <div>
    <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3">Direct child</span>
    <pre>
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    """

    def test_direct_sibling(self):
        """Test direct sibling."""

        # Spaces
        self.assert_selector(
            self.MARKUP,
            "span + span",
            ["5", "6"],
            flags=util.HTML
        )

    def test_direct_sibling_no_spaces(self):
        """Test direct sibling with no spaces."""

        # No spaces
        self.assert_selector(
            self.MARKUP,
            "span+span",
            ["5", "6"],
            flags=util.HTML
        )

    def test_complex_direct_siblings(self):
        """Test direct sibling with no spaces."""

        # Complex
        self.assert_selector(
            self.MARKUP,
            "span#\\34 + span#\\35",
            ["5"],
            flags=util.HTML
        )
